/* ────────────────────────────────────────────────────────────────────────── *
 │                                                                            │
 │ This file is part of the nbody-vectorization-toy-lab in the fram of the    │
 │ SPACE Center of Excellence                                                 │
 │ The work has been started by Luca Tornatore @ INAF (Italian National       │
 │ Institute for Astrophysics) and continued in collaboration with            │
 │ Vassilis Flouris @ FORTH (Foundation for Research and Technology - Hellas  │
 │                                                                            │
 │ Consult the accopanying file changes.log for details                       │
 │                                                                            │
 │ contact:                                                                   │
 │  luca.tornatore@inaf.it                                                    │
 │  vflouris@ics.forth.gr                                                     │
 │                                                                            │
 │                                                                            │
 │ ×××××××××××××××××××××××××××××××××××××××××××××××××××××××××××××××××××××××××  │
 │                                                                            │
 │                                                                            │
 * ────────────────────────────────────────────────────────────────────────── */


#if defined(USE_PAPI)                                           // -----------------------------------------------------------
#include <papi.h>


const unsigned int native_dp_scalar = (((0xc7 & 0xffffff)<<8) | (0xfc & 0xff));


#if defined(PAPI_MYLAPTOP)
#warning "using LOCAL events set"
#define PAPI_EVENTS_NUM 8
char     *papi_eventnames[PAPI_EVENTS_NUM] = {"Instructions", "cycles",
					      "L2D accesses", "L2D misses",
					      "SC DFP",
					      "SC SFP",
					      "V128 DFP",
					      "V256 DFP" };

char     *papi_namedevents[PAPI_EVENTS_NUM] = {"PAPI_TOT_INS","PAPI_TOT_CYC",
					       "PAPI_L2_DCA", "PAPI_L2_LDM",
					       "FP_ARITH:SCALAR_DOUBLE",
					       "FP_ARITH:SCALAR_SINGLE",
					       "FP_ARITH:128B_PACKED_DOUBLE",
					       "FP_ARITH:256B_PACKED_DOUBLE" };

#elif defined(PAPI_LEONARDO_DCGP)
#warning "using LEONARDO_DCGP events set"
#define PAPI_EVENTS_NUM 9
char     *papi_eventnames[PAPI_EVENTS_NUM] = {"Instructions", "cycles",
					      "L2D accesses", "L2D misses",
					      "SC DFP",
					      "SC SFP",
					      "V128 DFP",
					      "V256 DFP",
					      "V512 DFP" };

char     *papi_namedevents[PAPI_EVENTS_NUM] = {"PAPI_TOT_INS","PAPI_TOT_CYC",
					       "PAPI_L2_DCA", "PAPI_L2_LDM",
					       "FP_ARITH_INST_RETIRED:SCALAR_DOUBLE",
					       "FP_ARITH_INST_RETIRED:SCALAR_SINGLE",
					       "FP_ARITH_INST_RETIRED:128B_PACKED_DOUBLE",
					       "FP_ARITH_INST_RETIRED:256B_PACKED_DOUBLE",
					       "FP_ARITH_INST_RETIRED:512B_PACKED_DOUBLE" };

#elif defined(PAPI_LUMI_C)
#warning "using LUMI_C events set"
#define PAPI_EVENTS_NUM 7
char     *papi_eventnames[PAPI_EVENTS_NUM] = {"Instructions", "cycles",
					      "L2D accesses", "L2D misses",
					      "FP INS", "VFP INS", "FP OPS" };

char     *papi_namedevents[PAPI_EVENTS_NUM] = {"PAPI_TOT_INS","PAPI_TOT_CYC",
					       "PAPI_L2_DCA", "PAPI_L2_LDM",
					       "PAPI_FP_INS", "PAPI_VEC_INS", "PAPI_FP_OPS" };

#else

#error "please define the system at command line -DPAPI_MY_LAPTOP or -DPAPI_LEONARDO_DCGP"

#endif

int       papi_EventSet                = PAPI_NULL;             // the handle for the events' set
long long papi_buffer[PAPI_EVENTS_NUM] = {0};                   // storage for the counters' values
long long papi_values[PAPI_EVENTS_NUM] = {0};                   // accumulate the counters' values

                                                                // check that PAPI is OK, exit if not
#define PAPI_CHECK( R ) {						\
    if ( (R) != PAPI_OK ) {						\
      printf("a problem with PAPI (code %d) arise at line %d\n",	\
	     (R), __LINE__);fflush(stdout); return (R); }}


                                                                // check that PAPI is OK,
                                                                // issue a warning if not with a
                                                                // provided message
#define PAPI_WARN( R, S ) {						\
    if ( (R) != PAPI_OK ) {						\
      printf("a problem  with PAPI (code %d) arise at line %d: %s\n",	\
	     (R),  __LINE__, (S)); fflush(stdout); }}

                                                                // check that PAPI is OK about an event
                                                                // issue a warning if not with a
                                                                // provided message
#define PAPI_WARN_EVENT( R, E, S1, n ) {				\
    if ( (R) != PAPI_OK ) {						\
      printf("a problem  with PAPI (code %d) : event %d arise at line %d: %s (%d)\n", \
	     (R), (E),  __LINE__, (S1), (n)); fflush(stdout); }}

#define PAPI_WARN_NAMEDEVENT( R, E, S1, n ) {				\
    if ( (R) != PAPI_OK ) {						\
      printf("a problem  with PAPI (code %d) : event %s arise at line %d: %s (%d)\n", \
	     (R), (E),  __LINE__, (S1), (n)); fflush(stdout); }}


#define PAPI_ADD_EVENTS_to_SET { for ( int i = 0; i < PAPI_EVENTS_NUM; i++) { \
      retval = PAPI_query_event(papi_events[i]);			\
      if ( retval == PAPI_OK ) {					\
	retval = PAPI_add_event(papi_EventSet, papi_events[i]);		\
	PAPI_WARN_EVENT(retval, papi_events[i], "adding event", i);} else { \
      PAPI_WARN_EVENT(retval, papi_events[i],"querying event", i)}  } }

#define PAPI_ADD_NAMEDEVENTS_to_SET { for ( int i = 0; i < PAPI_EVENTS_NUM; i++) { \
      int native = 0;						\
      retval = PAPI_event_name_to_code(papi_namedevents[i] , &native );	\
      if ( retval == PAPI_OK ) {					\
	retval = PAPI_add_event(papi_EventSet, native);			\
	PAPI_WARN_NAMEDEVENT(retval, papi_namedevents[i], "adding event", i);} else { \
	PAPI_WARN_NAMEDEVENT(retval, papi_namedevents[i],"querying event", i)}  } }

#define PAPI_INIT {							\
    int retval = PAPI_library_init(PAPI_VER_CURRENT);			\
    if (retval != PAPI_VER_CURRENT)					\
      printf("wrong PAPI initialization: version %d instead of %d has been found\n", retval, PAPI_VER_CURRENT); \
    retval = PAPI_create_eventset(&papi_EventSet); PAPI_WARN(retval,"creating event set"); \
    PAPI_ADD_NAMEDEVENTS_to_SET; }
    //PAPI_ADD_EVENTS_to_SET; }

// to use HIGH-LEVEL API
//#define PAPI_START_CNTR { int res = PAPI_start_counters(papi_events, PAPI_EVENTS_NUM); PAPI_CHECK_RES(res); }
//#define PAPI_STOP_CNTR  { int res = PAPI_stop_counters(papi_values, PAPI_EVENTS_NUM); PAPI_CHECK_RES(res); }

// to use NORMAL API
#define PAPI_START_CNTR {						\
    int retval = PAPI_start(papi_EventSet); PAPI_WARN(retval, "starting counters"); }

#define PAPI_STOP_CNTR {						\
    memset(&papi_buffer[0], 0, sizeof(long long)*PAPI_EVENTS_NUM);	\
    int retval = PAPI_stop(papi_EventSet, papi_buffer);			\
    if( retval == PAPI_OK ) {						\
      for( int jj = 0; jj < PAPI_EVENTS_NUM; jj++)			\
	papi_values[jj] += papi_buffer[jj]; } else PAPI_WARN(retval, "reading counters"); }



#else                                                           // -----------------------------------------------------------
#define PAPI_EVENTS_NUM 0
#define PAPI_INIT
#define PAPI_START_CNTR
#define PAPI_STOP_CNTR

#endif                                                          // -----------------------------------------------------------
